/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE OggSQUISH PLAYER/CONVERTER SOURCE CODE. *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS SOURCE IS RESTRICTED  *
 * BY DISTRIBUTION TERMS THAT SHOULD BE INCLUDED WITH THIS SOURCE.  *
 * PLEASE READ THESE TERMS BEFORE USING OR DISTRIBUTING.            *
 *                                                                  *
 * THE OggSQUISH SOURCE CODE IS (C) COPYRIGHT 1994-1996             *
 * XIPHOPHORUS                                                      *
 *                                                                  *
 ********************************************************************

  $Header: /home/ogg/repository/ogg/apps/mainloop.c,v 1.3 1998/01/20 04:55:17 ogg Exp $
  $Author: ogg $
  $Date: 1998/01/20 04:55:17 $
  modifications by: Monty
  function: 3 stage processing loop

 ********************************************************************/

#ifndef lint
static char rcs_id[] = "$Id";
#endif

#include "oggapp.h"
#include "appio.h"
#include "header.h"
#include "oggstuff.h"
#include "setup.h"

long bytesin=0,bytesout=0;

extern int OGG_verbose;
extern int OGG_X;
extern int  OGG_ess;
extern long OGG_polyphonic;
extern long O_channels;
extern long O_rate;
extern long O_bits;
extern long O_mulaw;
extern long O_bigendian;
extern long O_signed;

extern long I_channels;
extern long I_rate;
extern long I_bits;
extern long I_bigendian;
extern long I_signed;

extern int OGG_quality;
extern int OGG_time;

extern long OGG_begin;
extern long OGG_end;
extern long OGG_frames;

extern int OGG_in;
extern int OGG_out;
extern int OGG_action;

extern int X_channels;
extern int X_resample;
extern int X_volume;

extern int X_any;

extern char *infilename, *outfilename;
extern FILE *infile;

#include "linux.c"
#include "sgi.c"
#include "sun.c"
#include "netbsd.c"

void ProcessLoop(long index){
  FILE *outfile;
  long begin,end,temp;
  char buffer[160];

  /* Actually open output file.  We didn't do this before in case the
     X browser is in use; we don't want to create spurious files */

  switch(OGG_out){
  case -1:
    /* Playback */
    break;
  default:
    /* file/bitstream output */

    if(outfilename==NULL || !strcmp(outfilename,"-"))
      outfile=stdout;
    else{
      outfile=fopen(outfilename,"wb");
      if(outfile==NULL)
	reporterror("Could not open output file.");
    }
  }

  static_buffer_shift(index);

  switch(OGG_out){
  case -1:
    /* playback */
    break;
  case 0:
    /* Do this in the section-by-section loop */
    break;
  case 1:
    /* Raw...  no header to write */
    break;
  case 2:
    /* WAVE; ship out a header */
    WriteWav(outfile,O_channels,O_rate,O_bits,&O_bigendian,&O_signed);
    break;
  case 3:
    /* AIFF-C; ship out a header */
    WriteAifc(outfile,O_channels,O_rate,O_bits,&O_bigendian,&O_signed);
    break;
  }

  report(" ");
  /* The processing loops *************************************/
  begin=OGG_begin;
  end=0;
  while(1){
    double hundredths=0,hunP=0;
    long bytesoutP=0,bytesinP=0,seconds=0,minutes=0,hours=0;
    int GLOBAL_finished=0;

    /* The outer loop handles Ogg input consisting of multiple
       sections */

    /* Our input is already ready to go; Set up possible transforms 
       and output */

    switch(OGG_out){
    case -1:
      setplayback(outfilename);
      break;
    case 0:
      script_setup(outfile,O_channels,OGG_polyphonic,O_rate,
		   O_bits,OGG_quality,OGG_time);
      break;
    case 1:case 2: case 3:
      /* when dealing with multisection Ogg input, if we're not
	 explicitly resampling/setting channels, etc, we may need to
	 force fit the new input to the old output because WAVE/AIFC
	 can't change format on the fly. */
      if(!X_resample && I_rate!=O_rate)X_resample=O_rate;
      if(!X_channels && I_channels!=O_channels)X_channels=O_channels;
      break;
    }

    if(X_any)
      transform_setup();

    /* Pipeline stages and decisions.... */
    /* stage 0 -> 1 data transfer: -1 static_buffer, 0 in Ogg, 1 = passbuffer 
       Use STAGE_sourse */

    while(1){
      /* Loop through audio data, a block at a time... */

      long Ibytesper=I_channels*((I_bits+7)>>3);
      long Obytesper=O_channels*((labs(O_bits)+7)>>3);
      int STAGE_source,STAGE_bytesper=Ibytesper;
      int STAGE_finished=0;
      void *pass_buffer=NULL;
      long pass_bytes=0;

      /* Input *****************************************/
      switch(OGG_in){
      case 0:
	/* Ogg input */
	STAGE_finished=input_stage_ogg(infile);

	pass_bytes=bytes_ogg_data(0,O_bits,O_channels);
	STAGE_bytesper=Obytesper;
	if(OGG_out==0 && I_bits!=O_bits){
	  STAGE_source=1;
	  pass_buffer=pull_ogg_data(0,O_bits,O_bigendian,O_signed);
	}else
	  STAGE_source=0;
	break;
      default:
	/* Raw sample; just grab some data */
	if(!feof(infile))static_buffer_read(infile,1024*Ibytesper);
	pass_bytes=static_buffer_size/Ibytesper*Ibytesper;
	STAGE_source=-1;
	if(feof(infile))STAGE_finished=-1;
      }

      /* Fast forwarding? (Well, skipping really) */

      if(begin){
	long remaining=STAGE_bytesper*(OGG_begin-begin);
	switch(STAGE_source){
	case -1:
	  if(remaining>pass_bytes)remaining=pass_bytes;
	  static_buffer_shift(remaining);
	  begin-=remaining/Ibytesper;
	  pass_bytes-=remaining;
	case 0:
	  if(remaining>=pass_bytes){
	    begin-=pass_bytes/Ibytesper;
	    pass_bytes=0;
	    break;
	  }

	  pass_buffer=pull_ogg_data(0,O_bits,O_bigendian,O_signed);
	  STAGE_source=1;
	  /* fall through on purpose */
	case 1:
	  
	  if(remaining>=pass_bytes){
	    begin-=pass_bytes/Ibytesper;
	    pass_bytes=0;
	    if(pass_buffer)free(pass_buffer);
	    break;
	  }

	  /* Hmmm.... Some will be left over.... */

	  memmove(pass_buffer,((char *)pass_buffer)+remaining,
		  pass_bytes-remaining);
	  pass_bytes-=remaining;
	  break;
	}
	sprintf(buffer,"\r=[  Fast Forwarding  ]==> %ldkB "
		"to go...   ",begin>>10);
      }

      if(begin<=0){

	/* Do we have an explicit stopping point? */
	if(OGG_frames!=-1){
	  long remaining=STAGE_bytesper*(OGG_frames-end);
	  
	  switch(STAGE_source){
	  case -1:
	    if(remaining>pass_bytes)
	      remaining=pass_bytes;
	    else
	      STAGE_finished=-1;
	    pass_bytes-=remaining;
	    end+=remaining/Ibytesper;
	    break;
	  case 0:
	    if(remaining<=pass_bytes){
	      pass_bytes-=remaining;
	      end+=remaining/Ibytesper;
	      STAGE_finished=-1;
	      break;
	    }
	    
	    pass_bytes=0;
	    break;
	  case 1:
	    if(remaining<pass_bytes){
	      pass_bytes-=remaining;
	      end+=remaining/Ibytesper;
	      STAGE_finished=-1;
	      break;
	    }
	    
	    if(pass_buffer)free(pass_buffer);
	    if(remaining==pass_bytes)STAGE_finished=-1;
	    pass_bytes=0;
	    break;
	  }
	}

	if(pass_bytes){
	  bytesin+=pass_bytes;

	  /* Any effects/transforms? */
	  if(X_any){
	    switch(STAGE_source){
	    case -1:
	      push_ogg_data(1,I_bits,I_bigendian,I_signed,static_buffer,
			    pass_bytes);
	      static_buffer_shift(pass_bytes);      
	      break;
	    case 0:
	      move_ogg_data(0,1);
	      break;
	    case 1:
	      push_ogg_data(1,O_bits,O_bigendian,O_signed,pass_buffer,
			    pass_bytes);
	      if(pass_buffer)free(pass_buffer);
	      break;
	    }
	    
	    /* Might have more than one block to process */
	    while(transform_stage_ogg(STAGE_finished)){
	      switch(OGG_out){
	      case 0:
		move_ogg_data(1,2);
		output_stage_ogg(outfile,STAGE_finished);
		break;
	      case -1:
		bytesout+=pass_bytes=bytes_ogg_data(1,O_bits,O_channels);
		pass_buffer=pull_ogg_data(1,O_bits,O_bigendian,O_signed);
		if(pass_bytes)playback(pass_buffer,pass_bytes,(O_bits+7)>>3);
#if !defined(HAVE_LIBPTHREAD) || !(defined(__linux__))
		free(pass_buffer);
#endif
      		break;
	      default:
		bytesout+=pass_bytes=bytes_ogg_data(1,O_bits,O_channels);
		pass_buffer=pull_ogg_data(1,O_bits,O_bigendian,O_signed);
		if(pass_bytes)blocking_write(pass_buffer,1,pass_bytes,outfile);
		free(pass_buffer);
		break;
	      }
	    }
	  }else{
	    switch(OGG_out){
	    case 0:
	      switch(STAGE_source){
	      case -1:
		push_ogg_data(2,I_bits,I_bigendian,I_signed,static_buffer,
			      pass_bytes);
		static_buffer_shift(pass_bytes);      
		break;
	      case 0:
		move_ogg_data(0,2);
		break;
	      case 1:
		push_ogg_data(2,O_bits,O_bigendian,O_signed,pass_buffer,
			      pass_bytes);
		if(pass_buffer)free(pass_buffer);
		break;
	      }
	      output_stage_ogg(outfile,STAGE_finished);
	      break;
	    case -1:
	      switch(STAGE_source){
	      case 0:
		pass_bytes=bytes_ogg_data(0,O_bits,O_channels);
		pass_buffer=pull_ogg_data(0,O_bits,O_bigendian,O_signed);
		/* fall through */
	      case 1:
		bytesout+=pass_bytes;
		if(pass_bytes)playback(pass_buffer,pass_bytes,(O_bits+7)>>3);
#if !defined(HAVE_LIBPTHREAD) || !(defined(__linux__))
		if(pass_buffer)free(pass_buffer);
#endif
		break;
	      }
	      break;
	    default:
	      switch(STAGE_source){
	      case 0:
		bytesout+=pass_bytes=bytes_ogg_data(0,O_bits,O_channels);
		pass_buffer=pull_ogg_data(0,O_bits,O_bigendian,O_signed);
		if(pass_bytes)blocking_write(pass_buffer,1,pass_bytes,outfile);
		if(pass_buffer)free(pass_buffer);
		break;
	      case 1:
		bytesout+=pass_bytes;
		if(pass_bytes)blocking_write(pass_buffer,1,pass_bytes,outfile);
		if(pass_buffer)free(pass_buffer);
		break;
	      }
	    }
	  }
	}
      }
      if(STAGE_finished)break;

      if(OGG_verbose){
	switch(OGG_out){
	case -1:

	  hundredths+=100.*(bytesout-bytesoutP)/Obytesper/O_rate;
	  if(hundredths-hunP>50){
	    flatreport("\r=[    Processing     ]==> ");
	    
	    temp=hundredths/100;
	    seconds+=temp;
	    hundredths-=temp*100;
	    hunP=hundredths;

	    temp=seconds/60;
	    minutes+=temp;
	    seconds-=temp*60;
	    
	    temp=minutes/60;
	    hours+=temp;
	    minutes-=temp*60;
	    
	    if(hours)
	      sprintf(buffer," %02ld:%02ld:%02ld.%02ld (%ldkB out)     ",
		      hours,minutes,seconds,(long)hundredths,bytesout>>10);
	    else
	      sprintf(buffer," %02ld:%02ld.%02ld (%ldkB out)     ",minutes,
		      seconds,(long)hundredths,bytesout>>10);
	    
	    flatreport(buffer);
	  }
	  bytesoutP=bytesout;
	  break;
	case 0:
	  if(bytesin-bytesinP>20*1024){
	    flatreport("\r=[    Processing     ]==> ");
	    sprintf(buffer,"%ldk",bytesin>>10);
	    flatreport(buffer);
	    sprintf(buffer," -> %ldk",bytesout>>10);
	    flatreport(buffer);
	    
	    if(bytesout-bytesoutP && bytesin>bytesout){
	      sprintf(buffer," (%#.3g:1",(double)bytesin/bytesout);
	      flatreport(buffer);
	      
	      sprintf(buffer,", IR %#.3g:1)        ",
		      (double)(bytesin-bytesinP)/(bytesout-bytesoutP));
	      bytesinP=bytesin;
	      bytesoutP=bytesout;
	      flatreport(buffer);
	    }
	  }
	  break;
	default:
	  if(bytesout-bytesoutP>20*1024){
	    flatreport("\r=[    Processing     ]==> ");
	    sprintf(buffer,"%ldkBytes out",bytesout>>10);
	    flatreport(buffer);
	    bytesoutP=bytesout;
	  }
	  break;
	}
      }
    }
	
    /* Inside cleanup */

    if(X_any){
      /* Nuke the transform script */
      ogg_run_set(1);
      ogg_run_reset();
    }

    switch(OGG_out){
    case -1:
      closeplayback();
      break;
    }

    switch(OGG_in){
    case 0:
      ogg_run_set(0);

      /* The static buffer at this point should be empty */
      {
	/* Recover buffered data */
	char *temp;
	long bread;
	ogg_output_bit(ogg_bit_in,ogg_primary,&temp,&bread);
	if(temp){
	  static_buffer_verify(bread);
	  memcpy(static_buffer,temp,bread);
	  static_buffer_size=bread;
	  free(temp);
	}
      }
      
      ogg_run_reset();

      /* This section is over....  Will we have another section? */
      
      if(feof(infile) && static_buffer_size==0){
	GLOBAL_finished=-1;
	break; /* Done! */
      }

      /* Hmmm....  look to see if we have another header. */
    
      flatreport("\r=[Loading new header ]=\n");
      CheckOgg(infile,&I_channels,&I_rate,&I_bits);
      if(ogg_errno){
	report("Garbage at end of file?\n");
	ogg_run_reset();
	GLOBAL_finished=-1;
	break;
      }

      sprintf(buffer,"Input changes loaded! %d bits, %d channel(s), %ld Hz",
	      (int)labs(I_bits),(int)I_channels,(long)I_rate);
      report(buffer);
    
      break;
    default:
      GLOBAL_finished=-1;
    }
    if(GLOBAL_finished)break;
  }

  /* Outside cleanup */
  
  fclose(infile);

  switch(OGG_out){
  case 0:case 1:
    /* Ogg is a stream model; no cleanup */
    fclose(outfile);
    break;
  case 2:
    FinishWav(outfile,O_channels,O_rate,O_bits);
    fclose(outfile);
    break;
  case 3:
    FinishAifc(outfile,O_channels,O_rate,O_bits);
    fclose(outfile);
    break;
  }

  infile = NULL;
  outfile = NULL;

  /***********************************************************/

  report("\r                                                                   ");
  switch(OGG_out){
  case 0:
    if((float)bytesin/bytesout>1.){
      sprintf(buffer,"\rDone.  Bytes out: %ld  Compression ratio: %#.3g:1\n\n",
	      bytesout,(float)bytesin/bytesout);
      report(buffer);
      break;
    }
    /* really*/
  default:
    sprintf(buffer,"\rDone.\n\n");
    report(buffer);
    break;
  }

  /* Clean up */
  if(static_buffer)free(static_buffer);
  static_buffer=NULL;
  static_buffer_alloced=0;
  static_buffer_size=0;
  /*if(antileak)free(antileak);*/
}
