/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE OggSQUISH PLAYER/CONVERTER SOURCE CODE. *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS SOURCE IS RESTRICTED  *
 * BY DISTRIBUTION TERMS THAT SHOULD BE INCLUDED WITH THIS SOURCE.  *
 * PLEASE READ THESE TERMS BEFORE USING OR DISTRIBUTING.            *
 *                                                                  *
 * THE OggSQUISH SOURCE CODE IS (C) COPYRIGHT 1994-1996             *
 * XIPHOPHORUS                                                      *
 *                                                                  *
 ********************************************************************

  file: sun.c
  function: Playback code for Sun Solaris
  author: Monty <xiphmont@mit.edu>
  modifications by: Monty
  last modification date: Oct 13 1996

 ********************************************************************/

/* Can we set a playback mode exactly matching the sample? */
/* If not, set the best we can, or use transforms and effects to force
   fit the playback to the hardware */

/* Sun drivers very similar to Linux/OSS (actually, it looks like
   Linux/OSS was based on the Sun interface, but Sun is cleaner) */

#if defined (__NetBSD__)

#include <sys/audioio.h>
#include <sys/ioctl.h>

int audio;

extern int OGG_ess;
extern long OGG_polyphonic;
extern long OGG_mulaw;
extern long O_channels;
extern long O_rate;
extern long O_bits;
extern long O_bigendian;
extern long O_signed;

extern long I_channels;
extern long I_rate;
extern long I_bits;
extern long I_bigendian;
extern long I_signed;

extern int OGG_quality;
extern int OGG_time;

extern long OGG_begin;
extern long OGG_end;
extern long OGG_frames;

extern int OGG_in;  /* 0 OggSquish, 1 Raw, 2 WAVE, 3 AIFF[-C] */
extern int OGG_out; /* -1 playback, 0 OggSquish, 1 Raw, 2 WAVE, 3 AIFF[-C] */
extern int OGG_action; /* 0 encode, 1 decode, 2 play */

extern int X_channel;
extern int X_resample;
extern int X_volume;

long blocksize;
long blockbytes;
char *blockbuffer=NULL;

void setplayback(char *outfilename){
  char buffer[160];
  audio_info_t a;

  if(outfilename==NULL){
    /* Try to open /dev/audio */
    audio=open("/dev/audio",O_WRONLY | O_EXCL);
  }else
    audio=open(outfilename,O_WRONLY | O_EXCL);

  if(audio==-1)
    reporterror("Unable to open audio playback device.");

  /* Set format */
  /* Go for exact first */
  AUDIO_INITINFO(&a);

  a.mode=AUMODE_PLAY;

  O_bigendian=OGG_SAMEENDIAN;
  switch(O_bits){
  case -8:
    a.play.encoding=AUDIO_ENCODING_ULAW;
    a.play.precision=8;
    break;
  case 8:
    a.play.encoding=AUDIO_ENCODING_PCM8;
    a.play.precision=8;
    O_signed=0;
    break;
  default:
    a.play.encoding=AUDIO_ENCODING_PCM16;
    a.play.precision=16;
    O_signed=-1;
    O_bits=16;
    break;
  }

  if(ioctl(audio,AUDIO_SETINFO,&a)<0){
    /* Oops, no good...  Let's try a bunch of formats till we can set one.
       Libogg can do these conversions trivially */
    /* First linear 16, then au8, then linear 8 */

    a.play.encoding=AUDIO_ENCODING_PCM16;
    a.play.precision=16;
    O_signed=-1;
    O_bits=16;
    if(ioctl(audio,AUDIO_SETINFO,&a)<0){
      a.play.encoding=AUDIO_ENCODING_PCM8;
      a.play.precision=8;
      O_bits=8;
      O_signed=0;
      if(ioctl(audio,AUDIO_SETINFO,&a)<0){
	a.play.encoding=AUDIO_ENCODING_ULAW;
	a.play.precision=8;
	O_bits=-8;
	if(ioctl(audio,AUDIO_SETINFO,&a)<0){
	  /* Oh, boy...  Die miserably... */
	  reporterror("Unable to set any common playback format");
	}else
	  report("Best available playback mode is 8 bit mulaw");
      }else
	report("Using 8 bit linear playback");
    }else
      report("Using 16 bit playback");
  }

  /* Well, OK, format is set... */
  
  /* Set channels */

  switch(O_channels){
  case 1:
    a.play.channels=1;
    if(ioctl(audio,AUDIO_SETINFO,&a)==-1)
      reporterror("Unable to set monophonic playback mode.");
    break;
  case 2:
    a.play.channels=2;
    if(ioctl(audio,AUDIO_SETINFO,&a)==-1){
      a.play.channels=1;
      O_channels=1;
      X_channels=1;
      if(ioctl(audio,AUDIO_SETINFO,&a)==-1)
	reporterror("Unable to set stereo or mono playback mode.");
      report("Stereo playback unavailable; using mono.");
    }
    break;
  default:
    a.play.channels=2;
    O_channels=2;
    X_channels=2;
    if(ioctl(audio,AUDIO_SETINFO,&a)<0){
      a.play.channels=1;
      O_channels=1;
      X_channels=1;
      if(ioctl(audio,AUDIO_SETINFO,&a)<0)
	reporterror("Unable to set stereo or mono playback mode.");
      report("Stereo playback unavailable; using mono.");
    }else
      report("Using stereo playback");
    break;
  }

  /* Now the really fun one; set the best rate we can... */
  /* Try the original rate first */

  a.blocksize=1024;

  {
    long rates[10]={8000,9600,11025,16000,18900,22050,32000,37800,44100,48000};
    long i,temp,flag=0;

    /* Didn't get the one we wanted; try other rates */
    /* First try up, then down */

    for(i=0;i<10;i++)if(rates[i]>=O_rate)break;
    temp=i;
    for(;i<10;i++){
      a.play.sample_rate=rates[i];
      if(ioctl(audio,AUDIO_SETINFO,&a)>=0){

	if(rates[i]!=O_rate){
	  sprintf(buffer,"Resampling to %ldHz playback rate",rates[i]);
	  report(buffer);
	  X_resample=O_rate=rates[i];
	}

	flag=-1;
	break;
      }
    }
    if(!flag)
      for(i=temp-1;i>=0;i--){
	a.play.sample_rate=rates[i];
	if(ioctl(audio,AUDIO_SETINFO,&a)>=0){

	  if(rates[i]!=O_rate){
	    sprintf(buffer,"Resampling to %ldHz playback rate",rates[i]);
	    report(buffer);
	    X_resample=O_rate=rates[i];
	  }

	  flag=-1;
	  break;
	}
      }
    if(!flag)
      reporterror("Unable to set any known playback speed");
  }

  ioctl(audio,AUDIO_GETINFO,&a);
  blocksize=a.blocksize;
  if(blockbuffer)free(blockbuffer);
  if(blocksize==0)
    reporterror("NetBSD requested a zero blocksize!");
  blockbuffer=malloc(blocksize);
  if(blockbuffer==NULL)
    reporterror("Unable to allocate playback buffer");
}

void closeplayback(void){
  close(audio);
  if(blockbuffer)free(blockbuffer);
  blockbuffer=NULL;
  blockbytes=0;
  blocksize=0;
}

void killplayback(void){
  ioctl(audio,AUDIO_FLUSH,NULL);
  closeplayback();
}

void playback(void *buffer,long bytes, long dummy){
  while(bytes){
    if(blockbytes){
      long b=bytes;
      if(bytes+blockbytes>blocksize)b=blocksize-blockbytes;
      memcpy(blockbuffer+blockbytes,buffer,b);
      buffer+=b;
      blockbytes+=b;
      bytes-=b;

      if(blockbytes>=blocksize){
	fd_blocking_write(blockbuffer,blocksize,audio);
	blockbytes=0;
      }
    }else{
      if(bytes>=blocksize){
	fd_blocking_write(buffer,blocksize,audio);
	buffer+=blocksize;
	bytes-=blocksize;
      }else{
	memcpy(blockbuffer,buffer,bytes);
	blockbytes+=bytes;
	return;
      }
    }
  }
}

#endif
