/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE OggSQUISH PLAYER/CONVERTER SOURCE CODE. *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS SOURCE IS RESTRICTED  *
 * BY DISTRIBUTION TERMS THAT SHOULD BE INCLUDED WITH THIS SOURCE.  *
 * PLEASE READ THESE TERMS BEFORE USING OR DISTRIBUTING.            *
 *                                                                  *
 * THE OggSQUISH SOURCE CODE IS (C) COPYRIGHT 1994-1997             *
 * XIPHOPHORUS                                                      *
 *                                                                  *
 ********************************************************************

  $Id: oggstuff.c,v 1.4 1997/12/27 06:39:55 ogg Exp $
  $Date: 1997/12/27 06:39:55 $
  author: Monty <xiphmont@mit.edu>
  modifications by: Monty, Wrong Mke
  function: Application libogg usage code

 ********************************************************************/

#include "oggapp.h"
#include "appio.h"
#include "header.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#ifndef lint
static char rcs_id[] = "$Id: oggstuff.c,v 1.4 1997/12/27 06:39:55 ogg Exp $";
#endif

extern long bytesin;
extern long bytesout;

static long OUTCOUNTER=0;

long CheckOgg(FILE *infile,long *channels,long *rate,long *bits){

  /* work in 1k blocks */

  /* normally, we wouldn't do it as below, but we have to keep bytes
     from the beginning if this turns out not to be an Ogg
     bitstream. */

  long bytes=0,read;
  while(1){

    /* Normally, we'd just send a new block, but it's more
       straightforward right now to resubmit from the beginning.  This
       means we have to reset the parser each time */

    ogg_script_reset();

    if(static_buffer_size<bytes)reporterror("Unexpected end of file!");
    /* tops off the static buffer */
    static_buffer_read(infile,bytes+=1024);

    /* Inspect the input; if it's an Ogg header, load it */
    read=ogg_script_loadheader(static_buffer,static_buffer_size);

    /* Check results */
    switch(ogg_errno){
    case 0:
      /* No error; did we finish? */
      if(ogg_script_ready){
	char buffer[160];
	struct ogg_dspinfo info;

	/* OK; libogg says it got a complete header.  Commit to
	   reading an ogg file and shift the static read buffer */
	/*static_buffer_shift(read);*/
	bytesin+=read;
	sprintf(buffer,"Input is an OggSQUISH version %d.%d (rev %d) "
		"bitstream.",ogg_bitstream_major,ogg_bitstream_minor,
		ogg_bitstream_revision);
	report(buffer);

	/* Send the script to the Ogg interpreter */
	ogg_script_use(0);

	switch(ogg_errno){
	case 0:
	  if(ogg_run_loaded && ogg_run_ready){

	    /* Query the primary dsp output buffer; this bitstream might
	       not have an audio component!  We also need the sample
	       information to report */
	    
	    if(!ogg_buffer_used(ogg_dsp_out,ogg_primary))
	      reporterror("Could not locate an audio component in the "
			  "bitstream.");
	    ogg_buffer_dsp(ogg_dsp_out,ogg_primary,&info);

	    /* This *really* shouldn't generate an error, but it's
               still good to check */
				
	    if(ogg_errno){
	      report("Problem checking OggSQUISH bitstream information:");
	      reporterror(ogg_strerr);
	    }
				
	    *channels=info.channels;
	    *rate=info.rate;
	    *bits=info.dynrange;

	    /* All done.  Successful header load. */
	    return(read);
	  }
	  
	  /* No errors, but ogg_run_ready is not set.  This
	     should never happen. */
	  reporterror("Internal error.");
	case 101:
	  /* Argh; the ogg bitstream header could not be built for
	     some reason.  This is probably a bug in the bitstream
	     decompression script in the header. */
	  /* Report the returned error */
	  reporterror(ogg_script_err_string);
	default:
	  /* Some other error; report it */
	  reporterror(ogg_strerr);
	}
      }
      
      /* No errors, but we didn't manage to read a whole header
	 either.  Try more data */

      break;
    case 7:
      /* Ah, OK.  This isn't an Ogg header */
      return(0);
    case 6:
      /* Oops; this version is too old */
      {
	char buffer[160];
	sprintf(buffer,"Input bitstream is from Ogg version %dalpha%d.",
		ogg_bitstream_major,ogg_bitstream_minor);
	report(buffer);
	reporterror("\nThe specified input file is an old OggSQUISH\n"
		    "bitstream that cannot be loaded by 0.98+ OggSQUISH;\n"
		    "bitstreams from before 0.98 are incompatible.");
      }
    case 8:
      reporterror("OggSQUISH input file header is corrupted.");
    case 100:
      /* This is probably due to a header requesting a feature we
         don't have */
      {
	char buffer[160];
	sprintf(buffer,"Input is an OggSQUISH version %d.%d (rev %d) "
		"bitstream.",ogg_bitstream_major,ogg_bitstream_minor,
		ogg_bitstream_revision);
	report(buffer);

	if(ogg_script_err_number==9)
	  if(MAJOR<ogg_bitstream_major || 
	     (MAJOR==ogg_bitstream_major && MINOR<ogg_bitstream_minor) ||
	     (MAJOR==ogg_bitstream_major && MINOR==ogg_bitstream_minor &&
	      REVISION<ogg_bitstream_revision))
	    reporterror("The input bitstream has requested a module "
			"function that seems to be newer than this version "
			"of OggSQUISH.  Consider upgrading to a more recent\n"
			"version of Ogg.");
	  else
	    reporterror("The input bitstream has requested a module "
			"function that this Ogg decoder does not \n"
			"understand.\n");
	report("Header corruption!");
	reporterror(ogg_script_err_string);
      }
    default:
      reporterror(ogg_strerr);
    }
  }
  /* never get here */
}

int input_stage_ogg(FILE *infile){
  ogg_run_set(0);
  if(ogg_errno)reporterror(ogg_strerr);

  while(ogg_input_query(ogg_bit_in,ogg_primary) && !feof(infile)){
    if(!feof(infile))static_buffer_read(infile,4096);
	
    /* Submit the grabbed bitstream to libogg */
	
    ogg_input_bit(ogg_bit_in,ogg_primary,static_buffer,
		      static_buffer_size);
	
    /* Check for errors */
    switch(ogg_errno){
    case 0:
      static_buffer_shift(static_buffer_size);
      break;
    default:
      reporterror(ogg_strerr);
    }
  }
      
  /* All set, call the engine on this frame */
  ogg_run();
  if(ogg_errno)reporterror(ogg_strerr);
      
  if(ogg_run_ready)return(0);
  return(-1);
}

long bytes_ogg_data(int bank,long bits,long channels){
  ogg_errno=0;
  ogg_run_set(bank);
  if(ogg_errno)return(0);
  bits=labs(bits);
  return(ogg_frames_dsp(bank,ogg_dsp_out,ogg_primary)*channels*((bits+7)>>3));
}

void *pull_ogg_data(int bank,long bits,long bigendian,long osigned){
  void *temp;
  long bytes;
  ogg_run_set(bank);
  if(ogg_errno)reporterror(ogg_strerr);

  {/******************************************************************/
    int i;
    for(i=1;i<10;i++)
      if(ogg_buffer_used(ogg_dsp_out,i)){
	char buffer[256];
	int j;
	struct ogg_dspinfo info;

	ogg_buffer_dsp(ogg_dsp_out,i,&info);
	bytes=0;
	ogg_output_dsp(ogg_dsp_out,i,(void **)&temp,
		       &bytes,32,OGG_SAMEENDIAN,-1,-1);

	if(ogg_errno)reporterror(ogg_strerr);
	for(j=0;j<info.channels;j++){
	  int k;
	  FILE *f;

	  sprintf(buffer,"ogg_%ld_%d-%d.out",OUTCOUNTER,i,j);
	  f=fopen(buffer,"w");
	  for(k=j;k<bytes;k+=info.channels)
	    fprintf(f,"%d\n",((size32 *)temp)[k]);
	  fclose(f);
	}
	free(temp);
      }
  }/******************************************************************/

  bytes=0;
  ogg_output_dsp(ogg_dsp_out,ogg_primary,(void **)&temp,
			 &bytes,bits,bigendian,osigned,-1);
  if(ogg_errno)reporterror(ogg_strerr);
  OUTCOUNTER++;
  return(temp);
}

void push_ogg_data(int bank,long bits,long bigendian,long osigned,
		   void *buffer,long bytes){

  long elements=bytes/((labs(bits)+7)>>3);
  ogg_run_set(bank);
  if(ogg_errno)reporterror(ogg_strerr);
  ogg_input_dsp(ogg_dsp_in,ogg_primary,static_buffer,elements,bits,
		bigendian,osigned,-1);

  /* Check for errors */
  if(ogg_errno)
    reporterror(ogg_strerr);
}

void move_ogg_data(int sourcebank,int destbank){
  ogg_run_set(sourcebank);
  if(ogg_errno)reporterror(ogg_strerr);
  ogg_move_dsp(destbank,ogg_dsp_in,ogg_primary,ogg_dsp_out,ogg_primary);
  if(ogg_errno)reporterror(ogg_strerr);
}

int transform_stage_ogg(int finished){
  ogg_run_set(1);
  if(ogg_errno)reporterror(ogg_strerr);  

  if(ogg_input_query(ogg_dsp_in,ogg_primary)==0 ||
	(finished && ogg_run_ready)){
    ogg_run();
    if(ogg_errno)reporterror(ogg_strerr);
    return(-1);
  }
  return(0);
}

void output_stage_ogg(FILE *out,int finished){
  ogg_run_set(2);
  if(ogg_errno)reporterror(ogg_strerr);  

  while(ogg_input_query(ogg_dsp_in,ogg_primary)==0 ||
	(finished && ogg_run_ready)){
    char *temp;
    long bytes;
    
    ogg_run();
    if(ogg_errno)reporterror(ogg_strerr);

    {/******************************************************************/
      int i;
      for(i=1;i<10;i++)
	if(ogg_buffer_used(ogg_dsp_out,i)){
	  char buffer[256];
	  int j;
	  struct ogg_dspinfo info;
	  
	  ogg_buffer_dsp(ogg_dsp_out,i,&info);
	  bytes=0;
	  ogg_output_dsp(ogg_dsp_out,i,(void **)&temp,
			 &bytes,32,OGG_SAMEENDIAN,-1,-1);
	  
	  if(ogg_errno)reporterror(ogg_strerr);
	  for(j=0;j<info.channels;j++){
	    int k;
	    FILE *f;
	    
	    sprintf(buffer,"ogg_%ld_%d-%d.out",OUTCOUNTER,i,j);
	    f=fopen(buffer,"w");
	    for(k=j;k<bytes;k+=info.channels)
	      fprintf(f,"%d\n",((size32 *)temp)[k]);
	    fclose(f);
	  }
	  free(temp);
	}
    }/******************************************************************/

    bytes=0;
    ogg_output_bit(ogg_bit_out,ogg_primary,&temp,
		   &bytes);
    if(ogg_errno)reporterror(ogg_strerr);
    
    if(temp){
      /* Write output */
      blocking_write(temp,1,bytes,out);
      bytesout+=bytes;
      free(temp);
    }
    OUTCOUNTER++;
  }
}
