/* Further subbanding/correlation twixt subbands will improve quality */
/* Adaptive bit allocation */


/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE OggSQUISH SOFTWARE CODEC SOURCE CODE.   *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS SOURCE IS RESTRICTED  *
 * BY DISTRIBUTION TERMS THAT SHOULD BE INCLUDED WITH THIS SOURCE.  *
 * PLEASE READ THESE TERMS BEFORE USING OR DISTRIBUTING.            *
 *                                                                  *
 * THE OggSQUISH SOURCE CODE IS (C) COPYRIGHT 1994-1996             *
 * XIPHOPHORUS                                                      *
 *                                                                  *
 ********************************************************************

  $Header: /home/ogg/repository/ogg/apps/setup.c,v 1.3 1998/01/03 19:58:56 ogg Exp $
  $Author: ogg $
  $Date: 1998/01/03 19:58:56 $
  modifications by: Monty
  function: Virtual machine setup and configuration

 ********************************************************************/

#include "oggapp.h"
#include "appio.h"

#ifndef lint
static char rcs_id[] = "$Id";
#endif

extern long bytesout;

void parse_errors(){
  char buffer[80];
  switch(ogg_errno){
  case 0:
    return;
  case 1: 
    reporterror("Out of memory."); 
  case 3:
    reporterror("Segfault; Out of memory?");
  case 100:case 101:
    report("");
    report("Internal error in script:");
    if(ogg_script_err_line!=-1)
      sprintf(buffer,"  line: %d",ogg_script_err_line);
    else
      sprintf(buffer,"  statement: %d",ogg_script_err_statement);
    report(buffer);
    sprintf(buffer,"  argument: %d",ogg_script_err_argument);
    report(buffer);
    reporterror(ogg_script_err_string);
  default:
    report("Internal error:");
    reporterror(ogg_strerr);
  }
}

extern long X_resample;
extern long X_channels;
extern long X_volume;

extern long O_channels;
extern long O_rate;
extern long O_bits;
extern long I_channels;
extern long I_rate;
extern long I_bits;

void standard_define(char *s,long val){
  char buffer[80];
  
  sprintf(buffer,"= %s %ld;",s,val);
  ogg_script_parse(buffer);
  parse_errors();
}

/* We don't want to be doing FFTs on vector lengths with large amounts
of factors > 4.... */

/* This returns a factor weight heuristic */

int tryfactor(long in){
  int ret=0;
  while(in>1){
    if(in%4==0){
      in/=4;
      ret+=4;
    }else{
      if(in%2==0){
	in/=2;
	ret+=4;
      }else{
	if(in%3==0){
	  in/=3;
	  ret+=7;
	}else{
	  long i=5;
	  for(;;i+=2)
	    if(in%i==0){
	      in/=i;
	      ret+=i*i;
	      break;
	    }
	}
      }
    }
  }
  return(ret);
}

int cheapest_within(long in, long percent){
  long plusminus=rint(in/100.*percent),i;
  long bestsofar=LONG_MAX,bestnum=in,temp;
  for(i=in-plusminus;i<=in+plusminus;i++){
    temp=tryfactor(i);
    if (i==in)temp--;
    if(temp<bestsofar){
      bestsofar=temp;
      bestnum=i;
    }
  }
  return(bestnum);
}

void transform_setup(void){
  long overlap=32;
  long hint=1024;
  standard_define("rate",I_rate);
  standard_define("channels",I_channels);
  standard_define("dyn",(O_bits==-8)?13:O_bits);
  standard_define("hint",hint);
  standard_define("orate",O_rate);
  standard_define("ochannels",O_channels);

  ogg_script_parse("Aid    ogg_dsp_in ogg_primary aP hint rate channels dyn;"
                   "Es     aP bP aP hint-overlap;"
                   "Spt    bP frameV;");
  parse_errors();

  if(X_channels){
    if(X_channels==1){
      int i;
      char buffer[200];
      for(i=0;i<I_channels;i++){
        sprintf(buffer,"Ecs    bP %d tempP;"
                "Eaa    bP tempP;",i);
        ogg_script_parse(buffer);
        parse_errors();
      }
      ogg_script_parse("Ecs    bP 0 bP;"
                       "EamF   bP 1 channels;");
      parse_errors();
    }else{
      int i;
      char buffer[200];
      ogg_script_parse("Spm    bP cP;"
                       "Ez     bP frameV channels 0;");
      parse_errors();
      for(i=0;i<I_channels && i<X_channels;i++){
        sprintf(buffer,"Ecs    cP %d tempP;"
                       "Ecr    bP %d tempP;",i,i);
        ogg_script_parse(buffer);
        parse_errors();
      }
    }
  }

  if(X_volume && X_volume!=100){
    standard_define("volume",X_volume);
    ogg_script_parse("EamF   bP volume 100;");
    parse_errors();
  }

  if(X_resample){
    standard_define("tweak",cheapest_within(hint*O_rate/I_rate,3));
    standard_define("overlap",overlap);

    if(I_rate>X_resample)
      ogg_script_parse("Je    restartV 1 re1L;"
                       "Ez    reoutP overlap channels 0;"
             "re1:      Ec    reoutP bP bP;"
                       "Spc   bP tempP;"
                       "Es    tempP tempP reoutP hint-overlap;"

                       "Ff    bP bP;"
                       "Es    bP bP cP tweak;"
                       "Ffis  bP bP 1 hint;"
                       "Whr   reinP overlap*tweak/hint;"
                       "Whl   bP overlap*tweak/hint;"
                       "Eaa   bP reinP;"
                       "Es    bP bP reinP tweak-overlap*tweak/hint;"
             "re2:      Sv    1 restartV;");
    else
      ogg_script_parse("Je    restartV 1 re1L;"
                       "Ez    reoutP overlap channels 0;"
             "re1:      Ec    reoutP bP bP;"
                       "Spc   bP tempP;"
                       "Es    tempP tempP reoutP hint-overlap;"

                       "Ff    bP bP;"
                       "Ez    cP tweak-hint channels 0;"
                       "Ec    bP cP bP;"

                       "Ffis  bP bP 1 hint;"
                       "Whr   reinP overlap*tweak/hint;"
                       "Whl   bP overlap*tweak/hint;"
                       "Eaa   bP reinP;"
                       "Es    bP bP reinP tweak-overlap*tweak/hint;"
             "re2:      Sv    1 restartV;");
    parse_errors();
  }
  
  

  ogg_script_parse("Aod    ogg_dsp_out ogg_primary bP orate ochannels dyn;"
                   "Shalt  frameV hint-overlap%");
  parse_errors();
  ogg_script_use(1);
  parse_errors();
}

void script_setup(FILE *outfile,int channels,int polyphonic,long rate,
		  int bits,int quality,int time){
  char *h;
  long bytes;
  
  switch(quality){
    /* Lossless and quasi-lossless are split out of the pack early */
  case 10:
    switch(time){
    case 0:

      /**********************************************************************/
      /* Uncompressed encoding                                              */
      /**********************************************************************/

      /* Just do the uncompressed case in one fell swoop. */
      standard_define("hint",1024);
#     include "oggUE.oh"
      parse_errors();
      ogg_script_use(2);
      parse_errors();
      
      /* Decode script is next */
      /* Our parser has been reset; gotta do the defines thing over again */
      standard_define("hint",1024);
#     include "oggUD.oh"
      parse_errors();
      
      /* Pack the script into an archive header */
      ogg_script_makeheader(&h,&bytes);
      parse_errors();
      
      /* Save that header! (Catch that rabbit?) */
      blocking_write(h, 1, bytes, outfile);
      bytesout+=bytes;
      
      return;
    default:
      /*********************************************************************/
      /* Lossless and quasi-lossless compression                           */
      /*********************************************************************/
      
      /* Time level      
	 1: Covhl only, no backend
	 2: Ddf
	 3: 2 pole no huff
	 4: 4 pole no huff
	 5: 4 pole, huff residue
	 6: + codeword negstream
	 7: + codeword maxstream
	 8: Covtl
	 9: ?
	 
	 */
      
      standard_define("hint",1024);    
      if(bits>16)
	ogg_script_parse("= ddr Dr64;");
      else
	ogg_script_parse("= ddr Dr32;");
      parse_errors();
      
      switch(time){
      case 3:
	ogg_script_parse("= coeffs 2;");
	break;
      case 4:case 5:case 6:case 7:case 8:case 9:
	ogg_script_parse("= coeffs 4;");
	break;
      }
      parse_errors();
      
      switch(time){
      case 1:
#       include "oggLE1.oh"
	break;
      case 2:
#       include "oggLE2.oh"
	break;
      case 3:case 4:
#       include "oggLE3.oh"
	break;
      case 5:
#       include "oggLE5.oh"
	break;
      case 6:
#       include "oggLE6.oh"
	break;
      case 7:
#       include "oggLE7.oh"
	break;
      case 8:case 9:
#       include "oggLE8.oh"
	break;
      }
      parse_errors();
      ogg_script_use(2);
      parse_errors();

      /* Decompression script is next *******************/
      
      standard_define("hint",1024);    
      if(bits>16)
	ogg_script_parse("= ddp Dp64;");
      else
	ogg_script_parse("= ddp Dp32;");
      switch(time){
      case 3:
	ogg_script_parse("= coeffs 2;");
	break;
      case 4:case 5:case 6:case 7:case 8:case 9:
	ogg_script_parse("= coeffs 4;");
	break;
      }
      parse_errors();
	
      switch(time){
      case 1:
#       include "oggLD1.oh"
	break;
      case 2:
#       include "oggLD2.oh"
	break;
      case 3:case 4:
#       include "oggLD3.oh"
	break;
      case 5:
#       include "oggLD5.oh"
	break;
      case 6:
#       include "oggLD6.oh"
	break;
      case 7:
#       include "oggLD7.oh"
	break;
      case 8:case 9:
#       include "oggLD8.oh"
	break;
      }
      parse_errors();

      /* Pack the script into an archive header */
      ogg_script_makeheader(&h,&bytes);
      parse_errors();
      
      /* Save that header! (Catch that rabbit?) */
      blocking_write(h, 1, bytes, outfile);
      bytesout+=bytes;
    
      
      return;
    }
  default:
#   include "oggYE1.oh"
    parse_errors();
    ogg_script_use(2);
    parse_errors();

#   include "oggYD1.oh"
    parse_errors();
      
    ogg_script_makeheader(&h,&bytes);
    parse_errors();
      
    blocking_write(h, 1, bytes, outfile);
    bytesout+=bytes;    
      
    return;
  }
}

