/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE OggSQUISH PLAYER/CONVERTER SOURCE CODE. *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS SOURCE IS RESTRICTED  *
 * BY DISTRIBUTION TERMS THAT SHOULD BE INCLUDED WITH THIS SOURCE.  *
 * PLEASE READ THESE TERMS BEFORE USING OR DISTRIBUTING.            *
 *                                                                  *
 * THE OggSQUISH SOURCE CODE IS (C) COPYRIGHT 1994-1996             *
 * XIPHOPHORUS                                                      *
 *                                                                  *
 ********************************************************************

  $Id: xogg.c,v 1.3 1997/12/27 05:39:59 ogg Exp $
  $Date: 1997/12/27 05:39:59 $
  author: Monty <xiphmont@mit.edu>
  modifications by: Monty, Wrong Mike
  function: xogg application top level

 ********************************************************************/

/* Most of the code that makes direct use of libogg is in oggstuff.c;
   setup.c organizes OggScripts and talks to the Ogg parser. */

#ifndef lint
static char rcs_id[] = "$Id: xogg.c,v 1.3 1997/12/27 05:39:59 ogg Exp $";
#endif

#include "xogg.h"
#include "appio.h"
#include "header.h"
#include "cmdparse.h"
#include "mainloop.h"

extern long bytes_indexed;

/* configuration globals */

int OGG_verbose=0;
int OGG_X=0;
int  OGG_ess=0;
long OGG_polyphonic=-1;
long O_channels=2;
long O_rate=44100;
long O_bits=16;
long O_mulaw=0;
long O_bigendian=OGG_LITTLEENDIAN;
long O_signed=-1;

long I_channels=2;
long I_rate=44100;
long I_bits=16;
long I_bigendian=OGG_LITTLEENDIAN;
long I_signed=-1;

int OGG_quality=10;
int OGG_time=7;

long OGG_begin=0;
long OGG_end=0;
long OGG_frames=-1;


int OGG_in=0;  /* 0 OggSquish, 1 Raw, 2 WAVE, 3 AIFF[-C] */
int OGG_out=2; /* -1 playback, 0 OggSquish, 1 Raw, 2 WAVE, 3 AIFF[-C] */
int OGG_action=0; /* 0 encode, 1 decode, 2 play */

int X_channels=0;
int X_resample=0;
int X_volume=0;

int X_any=0;

char *infilename=NULL, *outfilename=NULL;
FILE *infile=NULL;
long bytes_indexed;

int main(int argc,char *argv[]){
  signal(SIGPIPE,SIG_IGN);
  /* Reset some signal handlers to make sure that we restart
  interrupted system calls when appropriate/necessary. */
  signal(SIGSTOP,SIG_DFL);
  signal(SIGCONT,SIG_DFL);

  gtk_init(&argc,&argv);
  scan_command_line(argc,argv,-1);
  ProcessLoop(bytes_indexed);
  return(0);
}


const char OPTFORMAT[]="FARWB:E:N:hqvb!l!e!s!u!c:r:w!:X!:LMpQ:T:U***";

int scan_command_line(int argc,char *argv[],int probein){
  long i;
  long opnum;
  char opchar, *opstring, buffer[160];
  char *multi;
  char *action=NULL;
  int ret=0;

  if(argc>1)ret=1;
  OGG_verbose=checkoption(argc,argv,'v');
  if(checkoption(argc,argv,'h')){
    OGG_verbose=-1;
    OGG_X=0;
    report("\n"
	   "Usage: ogg encode [input_file] [output_file] [transforms]\n"
	   "                  [general/conversion/compression options]\n\n"
	   "       ogg decode [input_file] { -A/-R/-W } [output_file]\n"
	   "                  [transforms] [general/conversion options]\n\n"
	   "       ogg play [input_file] [transforms] [general/playback options]\n\n"
	   "       ogg [input_file] [options]\n\n"
	   "\n"
	   "General options:\n"
	   "  -B<fr>    Begin working at frame <fr>\n"
	   "  -E<fr>    End working at frame <fr>\n"
	   "  -h -help  Display this help information\n"
	   "  -N<fr>    Process <fr> frames of input\n"
	   "  -q        Query input file type/header and exit\n"
	   "  -v        Verbose operation\n"
	   "\n"
	   "Sample conversion options:\n"
	   "  (NOTE: conversion settings are normally unnecessary.\n"
	   "   These options override automatic settings and default\n"
	   "   settings for raw files.)\n"
	   "  -bI -bO   Force bigendian input/set bigendian output byte order\n"
	   "            (no effect on Ogg samples or WAVE/AIFC output)\n"
	   "  -c<ch>    Override input channels setting (no effect on Ogg\n"
	   "            input streams.\n"
	   "  -ess      Use a workaround that improves the poor 16 bit\n"
	   "            playback quality of the ESS AudioDrive chips\n"
	   "            (playback only)\n"
	   "  -lI -lO   Force littleendian input/set littleendian output (no\n"
	   "            effect on Ogg samples or WAVE/AIFC output)\n"
	   "  -r<fr>    Override input sample rate\n"
	   "  -sI -sO   Force signed input/set signed output data (no effect\n"
	   "            on Ogg samples or WAVE/AIFC output)\n"
	   "  -uI -uO   Force unsigned input/set unsigned output (no effect\n"
	   "            on Ogg samples or WAVE/AIFC output)\n"
	   "  -wI <b> \n"
	   "  -wO <b>   Force input/set output sample bit width\n"
	   "\n"
	   "Effects and transforms:\n"
	   "  -Xresample <freq>  Resample input to sample rate <freq> Hz\n"
	   "  -Xvolume <percent> Scale signal volume to <percent> of original\n"
	   "\n"
	   "Compression options:\n"
	   "  -F        fastest lossy compression\n"
	   "  -L        lossless compression\n"
	   "  -m        disable polyphonic analysis (multichannel only)\n"
	   "  -p        emable polyphonic analysis (default; multichannel only)\n"
	   "  -Q<qu>    Compression quality (1=low 9=high 10=lossless)\n"
	   "  -T<ti>    Compression time (1=fastest 8=smallest)\n"
	   "  -U        Create an uncompressed Ogg file\n\n"
	   "The 'encode' keyword is used to produce OggSquish bitstreams from\n"
	   "raw, RIFF-WAVE and AIFF[-C] files, as well as re-encode existing\n"
	   "OggSquish bitstreams.  The 'decode' option is used to produce\n"
	   "AIFF[-C], raw or RIFF-WAVE output files (using -A, -R and -W,\n"
	   "respectively) from an input file (the input need not be an\n"
	   "OggSquish bitstream). The 'play' option will play back an input\n"
	   "file directly (Ogg, AIFF, AIFC, WAVE or raw).\n\n"
	   "When one of 'encode', 'decode' or 'play' is not specified, ogg's\n"
	   "behavior depends on the input file type; if the input file is an\n"
	   "OggSquish bitstream, it will be played back; otherwise, the\n"
           "input file will be converted into an OggSquish bitstream.\n\n"
	   "The input and/or output file may be specified as '-', which\n"
	   "specifies stdin or stdout respectively. Not specifying an \n"
	   "output file will produce an output file with the base name of\n"
	   "the input file, plus the appropriate extention.\n");
    exit(0);
  }
  OGG_verbose=checkoption(argc,argv,'q')?-1:OGG_verbose;

  sprintf(buffer,"\nXiphophorus OggSQUISH"
	 " file converter/player (libogg v%d.%drev%d)\n(c) "
	 "1994-1997 Xiphophorus\n"
         " + LinuxThreads playback 7/97\n",
	 MAJOR,MINOR,REVISION);
  report(buffer);
	  
  /* first string keyword could be one of encode/decode/play */
  action=stringoption(argc,argv,1);
  if(action==NULL || (strcmp(action,"encode") && strcmp(action,"decode") && 
     strcmp(action,"play"))){
    /* No explicit action */
    infilename=action;
    action=NULL;
    outfilename=stringoption(argc,argv,2);
  }else{
    /* We have an explicit action */
    if(!strcmp(action,"encode")){
      OGG_action=0;
      OGG_out=0;
    }else
      if(!strcmp(action,"decode"))OGG_action=1;
      else
	if(!strcmp(action,"play")){
	  OGG_action=2;
	  OGG_out=-1;
	}else
	  reporterror("Unknown action specified");

    infilename=stringoption(argc,argv,2);
    outfilename=stringoption(argc,argv,3);
  }

  /* input file */ /* (later we should also accept URLs, input devices
     and such, for WWW and OggStream connects, etc */

  /* open the file */
  if(infilename==NULL || !strcmp(infilename,"-")){
    /* piped stdin input */
    infile=stdin;
  }else{
    /* regular file, it seems... */
    infile=fopen(infilename,"rb");
    if(infile==NULL)reporterror("Cannot open input file");
  }

  /* Check for file type.... */

  if(probein)
    bytes_indexed=QueryHeader(infile);
  else
    bytes_indexed=0;

  O_channels=I_channels;
  O_rate=I_rate;
  O_bits=I_bits;
  O_bigendian=I_bigendian;
  O_signed=I_signed;

  if(action==NULL){
    switch(OGG_in){
    case 0:
      OGG_action=2;
      OGG_out=-1;
      break;
    default:
      OGG_action=0;
      OGG_out=0;
    }
  }      

  sprintf(buffer,"%ld bits, %ld channel(s), %ld Hz",I_bits,I_channels,I_rate);
  flatreport(buffer);

  if(!OGG_in==0)
    if(I_signed)
      flatreport(", signed");
    else
      flatreport(", unsigned");

  if(!OGG_in==0)
    if(I_bigendian==OGG_BIGENDIAN)
      flatreport(", bigendian");
    else
      flatreport(", littleendian");
  flatreport("\n");

  if(checkoption(argc,argv,'q'))exit(0);
  
  /* The rest of the command line is options and numbers; we also
     might have an output file, catch that at the end. */
  
  i=0;
  while((opchar=optionnum(argc,argv,i++,&opnum,&multi,&opstring))){
    switch(opchar){
    case 'A': 
      if(OGG_action!=1)reporterror("'decode' action required for -A");
      OGG_out=3;
      report("Outputting AIFF-C audio file.");
      break;
    case 'R':
      if(OGG_action!=1)reporterror("'decode' action required for -R");
      OGG_out=1;
      report("Outputting raw audio file.");
      break;
    case 'W':
      if(OGG_action!=1)reporterror("'decode' action required for -W");
      OGG_out=2;
      report("Outputting RIFF/WAVE audio file.");
      break;
    case 'B': /* n */
      OGG_begin=opnum;
      break;
    case 'E':
      OGG_end=opnum;
      break;
    case 'N':
      OGG_frames=opnum;
      break;
    case 'b':
      switch(multi[0]){
      case 'I':
	if(OGG_in==0)
	  report("-bI option has no effect on OggSquish files");
	else{
	  I_bigendian=OGG_BIGENDIAN;
	  O_bigendian=OGG_BIGENDIAN;
	  report("Bigendian input specified.");
	}
	break;
      case 'O':
	if(OGG_out==0)
	  report("-bO option has no effect on OggSquish files");
	else{
	  O_bigendian=OGG_BIGENDIAN;
	  report("Bigendian output specified.");
	}
	break;
      default:
	sprintf(buffer,"Unknown option: -b%c",multi[0]);
	reporterror(buffer);
      }
      break;
    case 'l':
      switch(multi[0]){
      case 'I':
	if(OGG_in==0)
	  report("-lI option has no effect on OggSquish files");
	else{
	  I_bigendian=OGG_LITTLEENDIAN;
	  O_bigendian=OGG_LITTLEENDIAN;
	  report("Littleendian input specified.");
	}
	break;
      case 'O':
	if(OGG_out==0)
	  report("-lO option has no effect on OggSquish files");
	else{
	  O_bigendian=OGG_LITTLEENDIAN;
	  report("Littleendian output specified.");
	}
	break;
      default:
	sprintf(buffer,"Unknown option: -l%c",multi[0]);
	reporterror(buffer);
      }
      break;
    case 'c': /* n */
      if(OGG_in==0)
	report("-c has no effect when input is OggSquish.  Ignoring.");
      else{
	I_channels=opnum;
	O_channels=opnum;
	if(opnum>1)
	  report("Multichannel input override.");
	else
	  report("Single channel input override.");
      }
      break;
    case 'e': /* ..ss */
      if(strcmp(multi,"ss")){
	sprintf(buffer,"unknown option -e");
	strncat(buffer,multi,40);
	reporterror(buffer);
      }
      OGG_ess=-1;
      if(OGG_out!=-1)
	report("-ess is useful only during playback");
      else
	report("Tailoring playback to the ESS AudioDrive.");
      break;
    case 'r': /* n */
      sprintf(buffer,"Sampling rate override: %ld Hz",opnum);
      report(buffer);
      I_rate=opnum;
      O_rate=opnum;
      break;
    case 's':
      switch(multi[0]){
      case 'I':
	if(OGG_in==0)
	  report("-sI option has no effect on OggSquish files");
	else{
	  I_signed=-1;
	  O_signed=-1;
	  report("Signed input specified.");
	}
	break;
      case 'O':
	if(OGG_out==0)
	  report("-sO option has no effect on OggSquish files");
	else{
	  O_signed=-1;
	  report("Signed output specified.");
	}
	break;
      default:
	sprintf(buffer,"Unknown option: -s%c",multi[0]);
	reporterror(buffer);
      }
      break;
    case 'u':
      switch(multi[0]){
      case 'I':
	if(OGG_in==0)
	  report("-uI option has no effect on OggSquish files");
	else{
	  I_signed=0;
	  O_signed=0;
	  report("Unigned input specified.");
	}
	break;
      case 'O':
	if(OGG_out==0)
	  report("-uO option has no effect on OggSquish files");
	else{
	  O_signed=0;
	  report("Unigned output specified.");
	}
	break;
      default:
	sprintf(buffer,"Unknown option: -u%c",multi[0]);
	reporterror(buffer);
      }
      break;
    case 'w': /* n */
      switch(multi[0]){
      case 'I':
	if(OGG_in==0)
	  report("-wI option has no effect on OggSquish files");
	else{
	  if(opnum<1 || opnum>24)reporterror("Word bit size out of bounds.");
	  I_bits=opnum;
	  O_bits=opnum;
	  sprintf(buffer,"Input sample size override: %ld bits",O_bits);
	  report(buffer);
	}
	break;
      case 'O':
	if(OGG_out==0)
	  report("-wO option has no effect on OggSquish files");
	else{
	  if(opnum<1 || opnum>24)reporterror("Word bit size out of bounds.");
	  O_bits=opnum;
	  sprintf(buffer,"Output sample size override: %ld bits",O_bits);
	  report(buffer);
	}
	break;
      default:
	sprintf(buffer,"Unknown option: -w%c",multi[0]);
	reporterror(buffer);
      }
      break;
    case 'L':
      if(OGG_out!=0)report("-L only useful when outputting an Ogg bitstream");
      OGG_quality=10;
      report("Lossless compression output requested.");
      break;
    case 'F':
      if(OGG_out!=0)report("-F only useful when outputting an Ogg bitstream");
      OGG_time=1;
      if(OGG_quality==10)OGG_quality=5;
      report("Fast lossy compression output requested.");
      break;
    case 'm':
      if(OGG_out!=0)report("-m only useful when outputting an Ogg bitstream");
      OGG_polyphonic=0;
      report("Multichannel (non-polyphonic) analysis requested.");
      break;
    case 'p':
      if(OGG_out!=0)report("-p only useful when outputting an Ogg bitstream");
      OGG_polyphonic=-1;
      report("Polyphonic analysis requested.");
      break;
    case 'Q': /* n */
      if(OGG_out!=0)report("-Q only useful when outputting an Ogg bitstream");
      if(opnum<1 || opnum>10)reporterror("Quality setting must be between"
					 " 1 and 10 inclusive.");
      OGG_quality=opnum;
      if(OGG_quality==10)
	report("Lossless compression requested.");
      else{
	sprintf(buffer,"Compression quality setting: %ld.",opnum);
	report(buffer);
      }
      break;
    case 'T': /* n */
      if(OGG_out!=0)report("-T only useful when outputting an Ogg bitstream");
      if(opnum<1 || opnum>10)reporterror("Time setting must be between"
					 " 1 and 9 inclusive.");
      OGG_time=opnum;
      sprintf(buffer,"Compression time setting: %ld.",opnum);
      report(buffer);
      break;
    case 'U':
      if(OGG_out!=0)report("-U only useful when outputting an Ogg bitstream");
      OGG_time=0;
      OGG_quality=10;
      report("Uncompressed bitstream requested.");
      break;
    case 'X': /* ..resample ..volume ..channels*/
      if(!strcmp(multi,"resample")){
	if(opnum<200)reporterror("resample rate must be >200Hz");
	O_rate=X_resample=opnum;
	break;
      }
      if(!strcmp(multi,"volume")){
	if(opnum<1)reporterror("volume transform should be >0 percent");
	X_volume=opnum;
	break;
      }
      if(!strcmp(multi,"channels")){
	if(opnum<1)reporterror("channels transform must be at least 1");
	O_channels=X_channels=opnum;
	break;
      }
      sprintf(buffer,"Unknown transform/effect: ");
      strncat(buffer,multi,80);
      reporterror(buffer);
      break;
    case 'v':
      break;
    default:
      sprintf(buffer,"Unknown option: %c.",(char)opchar);
      reporterror(buffer);
      break;
    }
  }
  if(!OGG_frames && OGG_end)OGG_frames=OGG_end-OGG_begin;
  if(OGG_frames<0)OGG_frames=-1;
  if(O_channels<2)OGG_polyphonic=0;

  /* Input is already prepared; prepare output */

  switch(OGG_out){
  case -1:
    /* Playback */
    break;
  default:
    /* file/bitstream output */

    if(outfilename==NULL){
      /* output name was not supplied */
      /* Check da input */

      if(infile==stdin){
	/* Send output to STDOUT */

	if(OGG_out!=1){
	  reporterror("Outputting WAVE and AIFF[-C] requires a seekable "
		      "output stream.");
	}
      }else{
	char *temp=strrchr(infilename,'.');
	outfilename=calloc(strlen(infilename)+10,1);
	if(outfilename==NULL)reporterror("Couldn't alloc memory");
	if(temp==NULL)
	  strcpy(outfilename,infilename);
	else
	  strncpy(outfilename,infilename,temp-infilename);
	switch(OGG_out){
	case 0:
	  strcat(outfilename,".ogg");
	  break;
	case 1:
	  strcat(outfilename,".raw");
	  break;
	case 2:
	  strcat(outfilename,".wav");
	  break;
	case 3:
	  strcat(outfilename,".aifc");
	  break;
	}

	flatreport("No output file specified; writing to ");
	flatreport(outfilename);
	flatreport("\n");	
      }
    }
  }

  X_any=0;
  if(X_channels || X_volume || X_resample)X_any=-1;

  /* Dummy to do possibly necessary implicit conversions */
  if(OGG_in!=0 && OGG_out!=0)
    X_any=-1;


  return(ret);
}
