/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE OggSQUISH SOFTWARE CODEC SOURCE CODE.   *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS SOURCE IS RESTRICTED  *
 * BY DISTRIBUTION TERMS THAT SHOULD BE INCLUDED WITH THIS SOURCE.  *
 * PLEASE READ THESE TERMS BEFORE USING OR DISTRIBUTING.            *
 *                                                                  *
 * THE OggSQUISH SOURCE CODE IS (C) COPYRIGHT 1994-1996             *
 * XIPHOPHORUS                                                      *
 *                                                                  *
 ********************************************************************

  file: F.c
  function: Basic FFT and frequency domain transform toplevels
  author: Monty <xiphmont@mit.edu>
  modifications by: Monty
  last modification date: Jul 31 1996

 ********************************************************************/

#include <math.h>
#include "ogg.h"
#include "memory.h"
#include "parser.h"
#include "virtual.h"
#include "fft.h"
#include "ilog.h"
#include "F.h"

extern int OM_MACHINE;
extern int __ogg_precision;

/* This caches lookup information; for that reason, use a different
   transform module for each expected transform size */

typedef struct F{
  int ip;
  int op;

  long cachedsize;
  long *itrigcache;
  float *trigcache;
  int *splitcache;
  void (*itransform)(int,long *,long *,int *);
  void (*transform)(int,float *,float *,int *);
  void (*iinitcache)(int,long *,int *);
  void (*initcache)(int,float *,int *);

  long timerange;
  long scale;
  long scaled;
  int sizeoverride;
} F;

/* Fxx    pipeinP pipeoutP */

static void __ogga_F(module *m){
  F *f=(F *)m->state;
  void *b=__ogg_copyfromP(f->ip);
  long size=__ogg_sizeMB(b),i;
  int channels=__ogg_chanMB(b),ch;
  int scalen=f->sizeoverride?__ogg_fixedARG(m,2):1;
  int scaled=f->sizeoverride?__ogg_fixedARG(m,3):size;
  float *floatptr;

  if(scaled==0)scaled=1;
  scaled=f->scaled?scaled:1;

  if(size==0){
    __ogg_resetP(f->op);
    return;
  }
  if(f->cachedsize!=size){
    __ogg_free(f->trigcache);
    __ogg_free(f->itrigcache);
    __ogg_free(f->splitcache);
    f->cachedsize=size;

    /* Proper size for the largest cache requirements */

    f->trigcache=
      (float *)__ogg_calloc(3*size,sizeof(float),OM_MACHINE);

    if(f->iinitcache)
      f->itrigcache=(long *)__ogg_calloc(3*size,sizeof(long),OM_MACHINE);

    f->splitcache=(int *)__ogg_calloc(30,sizeof(int),OM_MACHINE);
    (*(f->initcache))(size,f->trigcache,f->splitcache);
    if(f->iinitcache)
      (*(f->iinitcache))(size,f->itrigcache,f->splitcache);
  }
  
  floatptr=__ogg_calloc(size,sizeof(float),OM_TEMP);
  for(ch=0;ch<channels;ch++){
    long *ptr=__ogg_ptrMB(b,ch);
    for(i=0;i<size;i++)floatptr[i]=ptr[i];
    (*(f->transform))(size,floatptr,f->trigcache,f->splitcache);
    if(f->scale){
      float temp=(float)scalen/(f->scale*scaled);
      for(i=0;i<size;i++)ptr[i]=rint(floatptr[i]*temp);
    }else
      for(i=0;i<size;i++)ptr[i]=rint(floatptr[i]);
  }
  __ogg_free(floatptr);

  __ogg_sendtoP(f->op,b);
}

void __oggc_Ff(module *m){
  F *f=(F *)(m->state=__ogg_calloc(1,sizeof(F),OM_MACHINE));
  f->ip=__ogg_fixedARG(m,0);
  f->op=__ogg_fixedARG(m,1);
  f->initcache=__ogg_fdrffti;
  f->transform=__ogg_fdrfftf;
  f->scale=0;
  m->action=__ogga_F;
}

void __oggc_Ffi(module *m){
  F *f=(F *)(m->state=__ogg_calloc(1,sizeof(F),OM_MACHINE));
  f->ip=__ogg_fixedARG(m,0);
  f->op=__ogg_fixedARG(m,1);
  f->initcache=__ogg_fdrffti;
  f->transform=__ogg_fdrfftb;
  f->scale=1;
  f->scaled=1;
  m->action=__ogga_F;
}

void __oggc_Fcq(module *m){
  F *f=(F *)(m->state=__ogg_calloc(1,sizeof(F),OM_MACHINE));
  f->ip=__ogg_fixedARG(m,0);
  f->op=__ogg_fixedARG(m,1);
  f->initcache=__ogg_fdcosqi;
  f->transform=__ogg_fdcosqf;
  f->scale=2;
  f->scaled=0;
  m->action=__ogga_F;
}

void __oggc_Fcqi(module *m){
  F *f=(F *)(m->state=__ogg_calloc(1,sizeof(F),OM_MACHINE));
  f->ip=__ogg_fixedARG(m,0);
  f->op=__ogg_fixedARG(m,1);
  f->initcache=__ogg_fdcosqi;
  f->transform=__ogg_fdcosqb;
  f->scale=2;
  f->scaled=1;
  m->action=__ogga_F;
}

void __oggc_Ffs(module *m){
  F *f=(F *)(m->state=__ogg_calloc(1,sizeof(F),OM_MACHINE));
  f->ip=__ogg_fixedARG(m,0);
  f->op=__ogg_fixedARG(m,1);
  f->sizeoverride=1;
  f->initcache=__ogg_fdrffti;
  f->transform=__ogg_fdrfftf;
  f->scale=1;
  f->scaled=1;
  m->action=__ogga_F;
}

void __oggc_Ffis(module *m){
  F *f=(F *)(m->state=__ogg_calloc(1,sizeof(F),OM_MACHINE));
  f->ip=__ogg_fixedARG(m,0);
  f->op=__ogg_fixedARG(m,1);
  f->sizeoverride=1;
  f->initcache=__ogg_fdrffti;
  f->transform=__ogg_fdrfftb;
  f->scale=1;
  f->scaled=1;
  m->action=__ogga_F;
}

void __oggc_Fcqs(module *m){
  F *f=(F *)(m->state=__ogg_calloc(1,sizeof(F),OM_MACHINE));
  f->ip=__ogg_fixedARG(m,0);
  f->op=__ogg_fixedARG(m,1);
  f->sizeoverride=1;
  f->initcache=__ogg_fdcosqi;
  f->transform=__ogg_fdcosqf;
  f->scale=2;
  f->scaled=1;
  m->action=__ogga_F;
}

void __oggc_Fcqis(module *m){
  F *f=(F *)(m->state=__ogg_calloc(1,sizeof(F),OM_MACHINE));
  f->ip=__ogg_fixedARG(m,0);
  f->op=__ogg_fixedARG(m,1);
  f->sizeoverride=1;
  f->initcache=__ogg_fdcosqi;
  f->transform=__ogg_fdcosqb;
  f->scale=2;
  f->scaled=1;
  m->action=__ogga_F;
}

static void __ogga_Fn(module *m){
  int p=__ogg_fixedARG(m,0);
  void *b=__ogg_copyfromP(p);
  long size=__ogg_sizeMB(b),i;
  int channels=__ogg_chanMB(b),ch;
  float temp=size?1/size:1;

  for(ch=0;ch<channels;ch++){
    long *ptr=__ogg_ptrMB(b,ch);
    for(i=0;i<size;i++)
      ptr[i]*=temp*(i+1);
  }

  __ogg_sendtoP(p,b);
}

void __oggc_Fn(module *m){
  m->action=__ogga_Fn;
}

#include "daub.h"

static void __ogga_Fwdaub(module *m){
  int inP=__ogg_fixedARG(m,0);
  long order=__ogg_fixedARG(m,1);

  void *b=__ogg_readfromP(inP);
  long size=__ogg_sizeMB(b),i;
  int channels=__ogg_chanMB(b),ch;

  if(size==0){
    __ogg_resetP(inP);
    return;
  }

  if(size>4){
    float *floatptr=__ogg_malloc(size*sizeof(float),OM_TEMP);
    float *workptr=__ogg_malloc(size*sizeof(float),OM_TEMP);
    float *result;
    
    for(ch=0;ch<channels;ch++){
      long *ptr=__ogg_ptrMB(b,ch);
      for(i=0;i<size;i++)floatptr[i]=ptr[i];
      switch(order){

      case 6:
	wt1(floatptr,workptr,size,1,&DAUB6_S);
	result=floatptr;
	break;
      case -6:
	wt1(floatptr,workptr,size,-1,&DAUB6_S);
	result=floatptr;
	break;
      case 12:
	wt1(floatptr,workptr,size,1,&DAUB12_S);
	result=floatptr;
	break;
      case -12:
	wt1(floatptr,workptr,size,-1,&DAUB12_S);
	result=floatptr;
	break;
      case 20:
	wt1(floatptr,workptr,size,1,&DAUB20_S);
	result=floatptr;
	break;
      case -20:
	wt1(floatptr,workptr,size,-1,&DAUB20_S);
	result=floatptr;
	break;
      case 4:
	result=__ogg_daub4f(floatptr,workptr,size);
	break;
      case -4:
	result=__ogg_daub4i(floatptr,workptr,size);
	break;
      }
      
      for(i=0;i<size;i++)ptr[i]=rint(result[i]);
    }
    __ogg_free(floatptr);
    __ogg_free(workptr);
  }

  __ogg_sendtoP(inP,b);
}

void __oggc_Fwdaub(module *m){
  m->action=__ogga_Fwdaub;
}

static void __ogga_Fwcoif(module *m){
  int inP=__ogg_fixedARG(m,0);
  long order=__ogg_fixedARG(m,1);

  void *b=__ogg_readfromP(inP);
  long size=__ogg_sizeMB(b),i;
  int channels=__ogg_chanMB(b),ch;

  if(size==0){
    __ogg_resetP(inP);
    return;
  }

  if(size>4){
    float *floatptr=__ogg_malloc(size*sizeof(float),OM_TEMP);
    float *workptr=__ogg_malloc(size*sizeof(float),OM_TEMP);
    float *result;
    
    for(ch=0;ch<channels;ch++){
      long *ptr=__ogg_ptrMB(b,ch);
      for(i=0;i<size;i++)floatptr[i]=ptr[i];
      switch(order){

      case 6:
	wt1(floatptr,workptr,size,1,&COIF6_S);
	result=floatptr;
	break;
      case -6:
	wt1(floatptr,workptr,size,-1,&COIF6_S);
	result=floatptr;
	break;
      case 12:
	wt1(floatptr,workptr,size,1,&COIF12_S);
	result=floatptr;
	break;
      case -12:
	wt1(floatptr,workptr,size,-1,&COIF12_S);
	result=floatptr;
	break;
      case 18:
	wt1(floatptr,workptr,size,1,&COIF18_S);
	result=floatptr;
	break;
      case -18:
	wt1(floatptr,workptr,size,-1,&COIF18_S);
	result=floatptr;
	break;

      }
      
      for(i=0;i<size;i++)ptr[i]=rint(result[i]);
    }
    __ogg_free(floatptr);
    __ogg_free(workptr);
  }

  __ogg_sendtoP(inP,b);
}

void __oggc_Fwcoif(module *m){
  m->action=__ogga_Fwcoif;
}
