/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE OggSQUISH SOFTWARE CODEC SOURCE CODE.   *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS SOURCE IS RESTRICTED  *
 * BY DISTRIBUTION TERMS THAT SHOULD BE INCLUDED WITH THIS SOURCE.  *
 * PLEASE READ THESE TERMS BEFORE USING OR DISTRIBUTING.            *
 *                                                                  *
 * THE OggSQUISH SOURCE CODE IS (C) COPYRIGHT 1994-1996             *
 * XIPHOPHORUS                                                      *
 *                                                                  *
 ********************************************************************

  file: P.c
  function: Encoder preprocessing
  author: Monty <xiphmont@mit.edu>
  modifications by: Monty
  last modification date: Oct 15 1996

 ********************************************************************/

#include "ogg.h"
#include "memory.h"
#include "parser.h"
#include "virtual.h"
#include "bitwise.h"
#include <limits.h>
#include <math.h>
#include "ilog.h"

static void __ogga_Povhl(module *m){
  int dataP=__ogg_fixedARG(m,0);
  int elsper=__ogg_fixedARG(m,1);
  int codeP=__ogg_fixedARG(m,2);

  void *buf=__ogg_readfromP(dataP);
  int insize=__ogg_sizeMB(buf);
  int chan=__ogg_chanMB(buf),ch;

  long spaces=(insize+elsper-1)/elsper;
  void *max=__ogg_allocMB(chan,spaces);

  for(ch=0;ch<chan;ch++){
    long *inptr=__ogg_ptrMB(buf,ch);
    long *maxptr=__ogg_ptrMB(max,ch);

    /* generate a spaced max vector */
    {
      long max,ptr,i=0,j;
      for(ptr=0;ptr<spaces;ptr++){
	max=0;
	for(j=0;j<elsper && i<insize;i++,j++)
	  if(labs(inptr[i])>max)max=labs(inptr[i]);
	maxptr[ptr]=ilog(max);
      }
    }
  }

  __ogg_appendtoP(dataP,buf);
  __ogg_sendtoP(codeP,max);
}

void __oggc_Povhl(module *m){
  m->action=__ogga_Povhl;
}

static void __ogga_Povh(module *m){
  int dataP=__ogg_fixedARG(m,0);
  int elsper=__ogg_fixedARG(m,1);
  int bitsper=__ogg_fixedARG(m,2);
  int codeP=__ogg_fixedARG(m,3);

  void *buf=__ogg_readfromP(dataP);
  int insize=__ogg_sizeMB(buf);
  int chan=__ogg_chanMB(buf),ch;

  long spaces=(insize+elsper-1)/elsper;
  void *max=__ogg_allocMB(chan,spaces);

  for(ch=0;ch<chan;ch++){
    long *inptr=__ogg_ptrMB(buf,ch);
    long *maxptr=__ogg_ptrMB(max,ch);

    /* generate a spaced max vector */
    {
      long max,ptr,i=0,j,shift;
      for(ptr=0;ptr<spaces;ptr++){
	max=0;
	for(j=0;j<elsper && i<insize;i++,j++)
	  if(labs(inptr[i])>max)max=labs(inptr[i]);
	shift=(maxptr[ptr]=ilog(max))-bitsper;
	if(shift<0)shift=0;
	i-=j;
	for(j=0;j<elsper && i<insize;i++,j++)
	  if(inptr[i]<0)
	    inptr[i]=-(-inptr[i]>>shift);
	  else
	    inptr[i]>>=shift;
      }
    }
  }

  __ogg_appendtoP(dataP,buf);
  __ogg_sendtoP(codeP,max);
}

void __oggc_Povh(module *m){
  m->action=__ogga_Povh;
}

static void __ogga_Pivh(module *m){
  int dataP=__ogg_fixedARG(m,0);
  int elsper=__ogg_fixedARG(m,1);
  int bitsper=__ogg_fixedARG(m,2);
  int codeP=__ogg_fixedARG(m,3);

  void *buf=__ogg_readfromP(dataP);
  int insize=__ogg_sizeMB(buf);
  int chan=__ogg_chanMB(buf),ch;

  long spaces=(insize+elsper-1)/elsper;
  void *max=__ogg_readfromP(codeP);
  long spaces2=__ogg_sizeMB(max);
  long chan2=__ogg_chanMB(max);

  for(ch=0;ch<chan && ch<chan2;ch++){
    long *inptr=__ogg_ptrMB(buf,ch);
    long *maxptr=__ogg_ptrMB(max,ch);

    {
      long max,ptr,i=0,j,shift;
      for(ptr=0;ptr<spaces && ptr<spaces2;ptr++){
	max=0;
	shift=maxptr[ptr]-bitsper;
	if(shift<0)shift=0;
	for(j=0;j<elsper && i<insize;i++,j++)
	    inptr[i]<<=shift;
      }
    }
  }

  __ogg_appendtoP(dataP,buf);
  __ogg_sendtoP(codeP,max);
}

void __oggc_Pivh(module *m){
  m->action=__ogga_Pivh;
}

